#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
class ObjectDict(dict):
    """ A dict subclass which allows access by named attribute.

    """
    __slots__ = ()

    def __getattr__(self, attr):
        if attr in self:
            return self[attr]
        msg = "'%s' object has no attribute '%s'"
        raise AttributeError(msg % (type(self).__name__, attr))



""" An extremely hack script that was used to render the dock guide icons.

"""
from enaml.qt.QtCore import *
from enaml.qt.QtGui import *


class GuidePad(object):

    LeftPosition = 0

    TopPosition = 1

    RightPosition = 2

    BottomPosition = 3

    CenterTop = 4

    SplitLeft = 5

    SplitTop = 6

    SplitRight = 7

    SplitBottom = 8

    SplitHorizontal = 9

    SplitVertical = 10

    CenterLeft = 11

    CenterRight = 12

    CenterBottom = 13

    @staticmethod
    def makePath(size):
        path = QPainterPath()
        rect = QRectF(0, 0, size.width(), size.height())
        path.addRoundedRect(rect, 2.0, 2.0)
        return path

    @staticmethod
    def makeTriPath():
        path = QPainterPath()
        path.moveTo(0.0, 0.0)
        path.lineTo(9.0, 0.0)
        path.lineTo(5.0, 4.0)
        path.lineTo(4.0, 4.0)
        path.lineTo(0.0, 0.0)
        return path

    def __init__(self, rect, position=None):
        self._rect = rect
        self._position = position
        self._opacity = 0.8
        self._path = GuidePad.makePath(rect.size())
        self._tri_path = GuidePad.makeTriPath()
        grad = QLinearGradient(0.0, 0.0, 0.0, 1.0)
        grad.setCoordinateMode(QGradient.ObjectBoundingMode)
        grad.setColorAt(0.0, QColor(0xF5, 0xF8, 0xFB))
        grad.setColorAt(0.33, QColor(0xF0, 0xF3, 0xF6))
        grad.setColorAt(0.66, QColor(0xE5, 0xE8, 0xEE))
        grad.setColorAt(1.0, QColor(0xDE, 0xE2, 0xE9))
        self._brush = QBrush(grad)
        grad = QLinearGradient(0.0, 0.0, 0.0, 1.0)
        grad.setCoordinateMode(QGradient.ObjectBoundingMode)
        grad.setColorAt(0.0, QColor(0xFC, 0xEC, 0xBE))
        grad.setColorAt(1.0, QColor(0xF7, 0xC7, 0x73))
        self._fill_brush = QBrush(grad)
        self._pen = QPen(QColor(0x8A, 0x91, 0x9C))

    def rect(self):
        return self._rect

    def setRect(self, rect):
        old = self._rect
        self._rect = rect
        if rect.isValid():
            if self._path is None or old.size() != rect.size():
                self._path = GuidePad.makePath(rect.size())
        else:
            self._path = None

    def contains(self, pos):
        return self._rect.contains(pos)

    def intersects(self, rect):
        return self._rect.intersects(rect)

    def guidePosition(self):
        return self._guide_pos

    def setGuidePosition(self, position):
        self._position = position

    def brush(self):
        return self._brush

    def setBrush(self, brush):
        self._brush = brush

    def pen(self):
        return self._pen

    def setPen(self, pen):
        self._pen = pen

    def opacity(self):
        return self._opacity

    def setOpacity(self, opacity):
        self._opacity = opacity

    def paint(self, painter):
        rect = self._rect
        if not rect.isValid():
            return
        painter.save()
        painter.translate(rect.x(), rect.y())

        # Draw the background
        painter.setOpacity(1.0)#self._opacity)
        painter.fillPath(self._path, self._brush)
        painter.setPen(self._pen)
        painter.drawPath(self._path)

        color = QColor(0x44, 0x58, 0x79)
        fill_brush = self._fill_brush
        painter.setPen(color)
        position = self._position
        if position == self.TopPosition:
            width = rect.width() - 8
            height = rect.height() / 2 - 4
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            painter.fillRect(QRect(5, 8, width - 1, height - 4), fill_brush)
            painter.setRenderHint(QPainter.Antialiasing)
            w = rect.width() / 2 + 5
            h = rect.height() - 5
            painter.translate(w, h)
            painter.rotate(180)
            painter.fillPath(self._tri_path, color)
        elif position == self.BottomPosition:
            width = rect.width() - 8
            height = rect.height() / 2 - 4
            painter.drawRect(QRect(4, height + 4, width, height))
            painter.fillRect(QRect(5, height + 5, width - 1, 3), color)
            painter.fillRect(QRect(5, height + 8, width - 1, height - 4), fill_brush)
            painter.setRenderHint(QPainter.Antialiasing)
            w = rect.width() / 2 - 4
            painter.translate(w, 6)
            painter.fillPath(self._tri_path, color)
        elif position == self.LeftPosition:
            width = rect.width() / 2 - 4
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            painter.fillRect(QRect(5, 8, width - 1, height - 4), fill_brush)
            painter.setRenderHint(QPainter.Antialiasing)
            w = rect.width() - 5
            h = rect.height() / 2 - 4
            painter.translate(w, h)
            painter.rotate(90)
            painter.fillPath(self._tri_path, color)
        elif position == self.RightPosition:
            width = rect.width() / 2 - 4
            height = rect.height() - 8
            painter.drawRect(QRect(width + 4, 4, width, height))
            painter.fillRect(QRect(width + 5, 5, width - 1, 3), color)
            painter.fillRect(QRect(width + 5, 8, width - 1, height - 4), fill_brush)
            painter.setRenderHint(QPainter.Antialiasing)
            h = rect.height() / 2 + 5
            painter.translate(6, h)
            painter.rotate(-90)
            painter.fillPath(self._tri_path, color)
        elif position == self.CenterTop:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            painter.fillRect(QRect(5, 8, width - 1, height - 4), fill_brush)
        elif position == self.CenterBottom:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5 + height - 4, width - 1, 3), color)
            painter.fillRect(QRect(5, 5, width - 1, height - 4), fill_brush)
        elif position == self.CenterLeft:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, 3, height), color)
            painter.fillRect(QRect(8, 5, width - 4, height - 1), fill_brush)
        elif position == self.CenterRight:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(width + 1, 5, 3, height - 1), color)
            painter.fillRect(QRect(5, 5, width - 4, height - 1), fill_brush)
        elif position == self.SplitTop:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            painter.fillRect(QRect(5, 8, width - 1, height / 2 - 2), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            painter.setPen(pen)
            painter.drawLine(5, 8 + height / 2 - 3, 5 + width - 1, 8 + height / 2 - 3)
        elif position == self.SplitBottom:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            h = height / 2 - 2
            painter.fillRect(QRect(5, 8 + h, width - 1, h), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            painter.setPen(pen)
            painter.drawLine(5, 8 + height / 2 - 2, 5 + width - 1, 8 + height / 2 - 2)
        elif position == self.SplitLeft:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            w = width / 2
            h = height - 4
            painter.fillRect(QRect(5, 8, w - 1, h), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            pen.setDashOffset(1)
            painter.setPen(pen)
            painter.drawLine(3 + w, 8, 3 + w, 8 + h)
        elif position == self.SplitRight:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            w = width / 2
            h = height - 4
            painter.fillRect(QRect(5 + w, 8, w - 1, h), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            pen.setDashOffset(1)
            painter.setPen(pen)
            painter.drawLine(5 + w, 8, 5 + w, 8 + h)
        elif position == self.SplitHorizontal:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            w = width / 4
            h = height - 4
            painter.fillRect(QRect(6 + w, 8, 2 * w - 1, h), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            pen.setDashOffset(1)
            painter.setPen(pen)
            painter.drawLine(6 + w, 8, 6 + w, 8 + h)
            painter.drawLine(4 + 3 * w, 8, 4 + 3 * w, 8 + h)
        elif position == self.SplitVertical:
            width = rect.width() - 8
            height = rect.height() - 8
            painter.drawRect(QRect(4, 4, width, height))
            painter.fillRect(QRect(5, 5, width - 1, 3), color)
            h = height / 4
            painter.fillRect(QRect(5, 8 + h, width - 1, 2 * h - 2), fill_brush)
            pen = QPen(color, 0, Qt.DotLine)
            pen.setDashPattern([1, 1])
            painter.setPen(pen)
            painter.drawLine(5, 8 + h, 4 + width, 8 + h)
            painter.drawLine(5, 10 + 2 * h, 4 + width, 10 + 2 * h)
        # Draw the indicator
        painter.restore()


def render_cross(painter):
    path = QPainterPath()
    path.moveTo(35.0, 0)
    path.lineTo(75.0, 0)
    path.lineTo(75.0, 25.0)
    path.lineTo(85.0, 35.0)
    path.lineTo(110.0, 35.0)
    path.lineTo(110.0, 75.0)
    path.lineTo(85.0, 75.0)
    path.lineTo(75.0, 85.0)
    path.lineTo(75.0, 110.0)
    path.lineTo(35.0, 110.0)
    path.lineTo(35.0, 85.0)
    path.lineTo(25.0, 75.0)
    path.lineTo(0.0, 75.0)
    path.lineTo(0.0, 35.0)
    path.lineTo(25.0, 35.0)
    path.lineTo(35.0, 25.0)
    path.lineTo(35.0, 0.0)
    painter.fillPath(path, QColor(0xFF, 0xFF, 0xFF, 0x99))
    painter.setPen(QPen(QColor(0x77, 0x77, 0x77), 1.0))
    painter.drawPath(path)


def render_cross_ex(painter):
    path = QPainterPath()
    path.moveTo(49.0, 0)
    path.lineTo(89.0, 0)
    path.lineTo(89.0, 39.0)
    path.lineTo(99.0, 49.0)
    path.lineTo(138.0, 49.0)
    path.lineTo(138.0, 89.0)
    path.lineTo(99.0, 89.0)
    path.lineTo(89.0, 99.0)
    path.lineTo(89.0, 138.0)
    path.lineTo(49.0, 138.0)
    path.lineTo(49.0, 99.0)
    path.lineTo(39.0, 89.0)
    path.lineTo(0.0, 89.0)
    path.lineTo(0.0, 49.0)
    path.lineTo(39.0, 49.0)
    path.lineTo(49.0, 39.0)
    path.lineTo(49.0, 0.0)
    painter.fillPath(path, QColor(0xFF, 0xFF, 0xFF, 0x99))
    painter.setPen(QPen(QColor(0x77, 0x77, 0x77), 1.0))
    painter.drawPath(path)


def render_north_cross(painter):
    path = QPainterPath()
    path.moveTo(35.0, 0)
    path.lineTo(75.0, 0)
    path.lineTo(75.0, 25.0)
    path.lineTo(85.0, 35.0)
    path.lineTo(110.0, 35.0)
    path.lineTo(110.0, 75.0)
    path.lineTo(85.0, 75.0)
    path.lineTo(75.0, 85.0)
    path.lineTo(75.0, 110.0)
    path.lineTo(35.0, 110.0)
    path.lineTo(35.0, 85.0)
    path.lineTo(25.0, 75.0)
    path.lineTo(0.0, 75.0)
    path.lineTo(0.0, 35.0)
    path.lineTo(25.0, 35.0)
    path.lineTo(35.0, 25.0)
    path.lineTo(35.0, 0.0)
    painter.fillPath(path, QColor(0xFF, 0xFF, 0xFF, 0x99))
    painter.setPen(QPen(QColor(0x77, 0x77, 0x77), 1.0))
    painter.drawPath(path)


def render_box(painter):
    path = QPainterPath()
    path.moveTo(0.0, 0.0)
    path.lineTo(40.0, 0.0)
    path.lineTo(40.0, 40.0)
    path.lineTo(0.0, 40.0)
    path.lineTo(0.0, 0.0)
    painter.fillPath(path, QColor(0xFF, 0xFF, 0xFF, 0x99))
    painter.setPen(QPen(QColor(0x77, 0x77, 0x77), 1.0))
    painter.drawPath(path)


def render_vbar(painter):
    path = QPainterPath()
    rect = QRectF(0, 0, 9, 30)
    path.addRoundedRect(rect, 2.0, 2.0)
    grad = QLinearGradient(0.0, 0.0, 0.0, 1.0)
    grad.setCoordinateMode(QGradient.ObjectBoundingMode)
    grad.setColorAt(0.0, QColor(0xF5, 0xF8, 0xFB))
    grad.setColorAt(0.33, QColor(0xF0, 0xF3, 0xF6))
    grad.setColorAt(0.66, QColor(0xE5, 0xE8, 0xEE))
    grad.setColorAt(1.0, QColor(0xDE, 0xE2, 0xE9))
    brush = QBrush(grad)
    pen = QPen(QColor(0x8A, 0x91, 0x9C))

    painter.fillPath(path, brush)
    painter.setPen(pen)
    painter.drawPath(path)

    color = QColor(0x44, 0x58, 0x79)
    painter.fillRect(QRect(4, 4, 2, 23), color)


def render_hbar(painter):
    path = QPainterPath()
    rect = QRectF(0, 0, 30, 9)
    path.addRoundedRect(rect, 2.0, 2.0)
    grad = QLinearGradient(0.0, 0.0, 0.0, 1.0)
    grad.setCoordinateMode(QGradient.ObjectBoundingMode)
    grad.setColorAt(0.0, QColor(0xF5, 0xF8, 0xFB))
    grad.setColorAt(0.33, QColor(0xF0, 0xF3, 0xF6))
    grad.setColorAt(0.66, QColor(0xE5, 0xE8, 0xEE))
    grad.setColorAt(1.0, QColor(0xDE, 0xE2, 0xE9))
    brush = QBrush(grad)
    pen = QPen(QColor(0x8A, 0x91, 0x9C))

    painter.fillPath(path, brush)
    painter.setPen(pen)
    painter.drawPath(path)

    color = QColor(0x44, 0x58, 0x79)
    painter.setPen(color)
    painter.fillRect(QRect(4, 4, 23, 2), color)


def render_background(painter):
    brush = QBrush(QColor(0x00, 0x00, 0x00, 0x10), Qt.Dense6Pattern)
    painter.fillRect(QRect(0, 0, 129, 129), brush)
    brush = QBrush(QColor(0xFF, 0xFF, 0xFF, 0x10), Qt.Dense6Pattern)
    painter.translate(0, 1)
    painter.fillRect(QRect(0, 0, 129, 129), brush)


app = QApplication([])
image = QImage(QSize(128, 128), QImage.Format_ARGB32_Premultiplied)
image.fill(0)
painter = QPainter(image)
#render_box(painter)
#render_cross(painter)
#render_cross_ex(painter)
#render_vbar(painter)
#render_hbar(painter)
render_background(painter)
#pad = GuidePad(QRect(0, 0, 30, 30), GuidePad.CenterQuads)
#pad.paint(painter)
painter.end()

import os
path = os.path.join(os.path.dirname(__file__), 'background.png')
image.save(path)



#------------------------------------------------------------------------------
# Copyright (c) 2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------

AUTUMN_THEME = {
    "python": {
        "comment": {
            "color": "#aaaaaa",
            "font-style": "italic"
        },
        "single_quoted_string": {
            "color": "#aa5500"
        },
        "class_name": {
            "color": "#00aa00",
            "text-decoration": "underline"
        },
        "function_method_name": {
            "color": "#00aa00"
        },
        "operator": {
            "paper": "#ffffff"
        },
        "double_quoted_string": {
            "color": "#aa5500"
        },
        "triple_double_quoted_string": {
            "color": "#aa5500"
        },
        "decorator": {
            "color": "#888888"
        },
        "comment_block": {
            "color": "#aaaaaa",
            "font-style": "italic"
        },
        "keyword": {
            "color": "#0000aa"
        },
        "triple_single_quoted_string": {
            "color": "#aa5500"
        },
        "unclosed_string": {
            "color": "#FF0000",
            "paper": "#FFAAAA"
        },
        "highlighted_identifier": {
            "paper": "#ffffcc"
        }
    },
    "settings": {
        "caret": "#000000",
        "color": "#000000",
        "paper": "#ffffff",
        "name": "autumn"
    }
}

AUTUMN_THEME['enaml'] = AUTUMN_THEME['python']



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from enaml.qt.QtCore import Qt, QRect, QSize, QPoint, QTimer, Signal
from enaml.qt.QtWidgets import QApplication, QFrame, QLayout

from .event_types import (
    QDockItemEvent, DockItemShown, DockItemHidden, DockItemClosed
)
from .q_dock_tab_widget import QDockTabWidget
from .q_dock_title_bar import QDockTitleBar
from .utils import repolish


class _AlertData(object):
    """ A private class which stores the data needed for item alerts.

    """
    def __init__(self, timer, level, on, off, repeat, persist):
        self.timer = timer
        self.level = level
        self.on = on
        self.off = off
        self.repeat = repeat
        self.persist = persist
        self.remaining = repeat
        self.active = False


class QDockItemLayout(QLayout):
    """ A QLayout subclass for laying out a dock item.

    """
    def __init__(self, parent=None):
        """ Initialize a QDockAreaLayout.

        Parameters
        ----------
        parent : QWidget or None
            The parent widget owner of the layout.

        """
        super(QDockItemLayout, self).__init__(parent)
        self._size_hint = QSize()
        self._min_size = QSize()
        self._max_size = QSize()
        self._title_bar = None
        self._dock_widget = None

    #--------------------------------------------------------------------------
    # Public API
    #--------------------------------------------------------------------------
    def titleBarWidget(self):
        """ Get the title bar widget set for the layout.

        Returns
        -------
        result : IDockItemTitleBar or None
            The title bar widget for the layout, or None if no widget
            is applied.

        """
        return self._title_bar

    def setTitleBarWidget(self, title_bar):
        """ Set the title bar widget for the layout.

        The old widget will be hidden and unparented, but not destroyed.

        Parameters
        ----------
        title_bar : IDockItemTitleBar or None
            A concrete implementor of the title bar interface, or None.

        """
        old_bar = self._title_bar
        if old_bar is not None:
            old_bar.hide()
            old_bar.setParent(None)
        self._title_bar = title_bar
        if title_bar is not None:
            title_bar.setParent(self.parentWidget())
        self.invalidate()

    def dockWidget(self):
        """ Get the dock widget set for the layout.

        Returns
        -------
        result : QWidget
            The primary widget set in the dock item layout.

        """
        return self._dock_widget

    def setDockWidget(self, widget):
        """ Set the dock widget for the layout.

        The old widget will be hidden and unparented, but not destroyed.

        Parameters
        ----------
        widget : QWidget
            The widget to use as the primary content in the layout.

        """
        old_widget = self._dock_widget
        if widget is old_widget:
            return
        if old_widget is not None:
            old_widget.hide()
            old_widget.setParent(None)
        self._dock_widget = widget
        if widget is not None:
            widget.setParent(self.parentWidget())
        self.invalidate()

    #--------------------------------------------------------------------------
    # QLayout API
    #--------------------------------------------------------------------------
    def invalidate(self):
        """ Invalidate the layout.

        """
        super(QDockItemLayout, self).invalidate()
        self._size_hint = QSize()
        self._min_size = QSize()
        self._max_size = QSize()

    def setGeometry(self, rect):
        """ Set the geometry for the items in the layout.

        """
        super(QDockItemLayout, self).setGeometry(rect)
        title = self._title_bar
        widget = self._dock_widget
        title_rect = QRect(rect)
        widget_rect = QRect(rect)
        if title is not None and not title.isHidden():
            msh = title.minimumSizeHint()
            title_rect.setHeight(msh.height())
            widget_rect.setTop(title_rect.bottom() + 1)
            title.setGeometry(title_rect)
        if widget is not None and not widget.isHidden():
            widget.setGeometry(widget_rect)

    def sizeHint(self):
        """ Get the size hint for the layout.

        """
        sh = self._size_hint
        if not sh.isValid():
            width = height = 0
            title = self._title_bar
            widget = self._dock_widget
            if title is not None and not title.isHidden():
                hint = title.sizeHint()
                width += hint.width()
                height += hint.height()
            if widget is not None and not widget.isHidden():
                hint = widget.sizeHint()
                width = max(width, hint.width())
                height += hint.height()
            sh = self._size_hint = QSize(width, height)
        return sh

    def minimumSize(self):
        """ Get the minimum size for the layout.

        """
        ms = self._min_size
        if not ms.isValid():
            width = height = 0
            title = self._title_bar
            widget = self._dock_widget
            if title is not None and not title.isHidden():
                hint = title.minimumSizeHint()
                width += hint.width()
                height += hint.height()
            if widget is not None and not widget.isHidden():
                hint = widget.minimumSizeHint()
                width = max(width, hint.width())
                height += hint.height()
            ms = self._min_size = QSize(width, height)
        return ms

    def maximumSize(self):
        """ Get the maximum size for the layout.

        """
        ms = self._max_size
        if not ms.isValid():
            widget = self._dock_widget
            parent = self.parentWidget()
            if widget is not None and parent.isFloating():
                ms = widget.maximumSize()
                title = self._title_bar
                if title is not None and not title.isHidden():
                    height = ms.height() + title.minimumSizeHint().height()
                    ms.setHeight(min(16777215, height))
            else:
                ms = QSize(16777215, 16777215)
            self._max_size = ms
        return ms

    #--------------------------------------------------------------------------
    # QLayout Abstract API
    #--------------------------------------------------------------------------
    def addItem(self, item):
        """ A required virtual method implementation.

        """
        msg = 'Use `setTitleBarWidget | setDockWidget` instead.'
        raise NotImplementedError(msg)

    def count(self):
        """ A required virtual method implementation.

        This method should not be used and returns a constant value.

        """
        return 0

    def itemAt(self, idx):
        """ A virtual method implementation which returns None.

        """
        return None

    def takeAt(self, idx):
        """ A virtual method implementation which does nothing.

        """
        return None


class QDockItem(QFrame):
    """ A QFrame subclass which acts as an item QDockArea.

    """
    #: A signal emitted when the maximize button is clicked. This
    #: signal is proxied from the current dock item title bar.
    maximizeButtonClicked = Signal(bool)

    #: A signal emitted when the restore button is clicked. This
    #: signal is proxied from the current dock item title bar.
    restoreButtonClicked = Signal(bool)

    #: A signal emitted when the close button is clicked. This
    #: signal is proxied from the current dock item title bar.
    closeButtonClicked = Signal(bool)

    #: A signal emitted when the link button is toggled. This
    #: signal is proxied from the current dock item title bar.
    linkButtonToggled = Signal(bool)

    #: A signal emitted when the pin button is toggled. This
    #: signal is proxied from the current dock item title bar.
    pinButtonToggled = Signal(bool)

    #: A signal emitted when the title is edited by the user. This
    #: signal is proxied from the current dock item title bar.
    titleEdited = Signal(str)

    #: A signal emitted when the empty area is left double clicked.
    #: This signal is proxied from the current dock item title bar.
    titleBarLeftDoubleClicked = Signal(QPoint)

    #: A signal emitted when the empty area is right clicked. This
    #: signal is proxied from the current dock item title bar.
    titleBarRightClicked = Signal(QPoint)

    #: A signal emitted when the item is alerted. The payload is the
    #: new alert level. An empty string indicates no alert.
    alerted = Signal(str)

    def __init__(self, parent=None):
        """ Initialize a QDockItem.

        Parameters
        ----------
        parent : QWidget, optional
            The parent of the dock item.

        """
        super(QDockItem, self).__init__(parent)
        layout = QDockItemLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.setLayout(layout)
        self.setTitleBarWidget(QDockTitleBar())
        self.alerted.connect(self._onAlerted)
        self._manager = None  # Set and cleared by the DockManager
        self._alert_data = None
        self._vis_changed = None
        self._closable = True
        self._closing = False

    #--------------------------------------------------------------------------
    # Reimplementations
    #--------------------------------------------------------------------------
    def close(self):
        """ Handle the close request for the dock item.

        """
        self._closing = True
        try:
            super(QDockItem, self).close()
        finally:
            self._closing = False

    def closeEvent(self, event):
        """ Handle the close event for the dock item.

        This handler will reject the event if the item is not closable.

        """
        event.ignore()
        if self._closable:
            event.accept()
            area = self.rootDockArea()
            if area is not None and area.dockEventsEnabled():
                event = QDockItemEvent(DockItemClosed, self.objectName())
                QApplication.postEvent(area, event)

    def showEvent(self, event):
        """ Handle the show event for the container.

        This handler posts a visibility change event.

        """
        super(QDockItem, self).showEvent(event)
        self._postVisibilityChange(True)

    def hideEvent(self, event):
        """ Handle the hide event for the container.

        This handler posts a visibility change event.

        """
        super(QDockItem, self).hideEvent(event)
        # Don't post when closing; A closed event is posted instead.
        if not self._closing:
            self._postVisibilityChange(False)

    def mousePressEvent(self, event):
        """ Handle the mouse press event for the dock item.

        This handler will clear any alert level on a left click.

        """
        if event.button() == Qt.LeftButton:
            self.clearAlert()
        super(QDockItem, self).mousePressEvent(event)

    #--------------------------------------------------------------------------
    # Public API
    #--------------------------------------------------------------------------
    def manager(self):
        """ Get the dock manager for this dock item.

        Returns
        -------
        result : DockManager or None
            The dock manager which is managing this item.

        """
        return self._manager

    def rootDockArea(self):
        """ Get the root dock area for this dock item.

        Returns
        -------
        result : QDockArea or None
            The root dock area for this dock item.

        """
        manager = self._manager
        if manager is not None:
            return manager.dock_area()

    def title(self):
        """ Get the title for the dock item.

        Returns
        -------
        result : unicode
            The unicode title for the dock item.

        """
        return self.titleBarWidget().title()

    def setTitle(self, title):
        """ Set the title for the dock item.

        Parameters
        ----------
        title : unicode
            The unicode title to use for the dock item.

        """
        self.titleBarWidget().setTitle(title)
        # A concession to practicality: walk the ancestry and update
        # the tab title if this item lives in a dock tab.
        container = self.parent()
        if container is not None:
            stacked = container.parent()
            if stacked is not None:
                tabs = stacked.parent()
                if isinstance(tabs, QDockTabWidget):
                    index = tabs.indexOf(container)
                    tabs.setTabText(index, title)

    def icon(self):
        """ Get the icon for the dock item.

        Returns
        -------
        result : QIcon
            The icon in use for the dock item.

        """
        return self.titleBarWidget().icon()

    def setIcon(self, icon):
        """ Set the icon for the dock item.

        Parameters
        ----------
        icon : QIcon
            The icon to use for the dock item.

        """
        self.titleBarWidget().setIcon(icon)
        # A concession to practicality: walk the ancestry and update
        # the tab icon if this item lives in a dock tab.
        container = self.parent()
        if container is not None:
            stacked = container.parent()
            if stacked is not None:
                tabs = stacked.parent()
                if isinstance(tabs, QDockTabWidget):
                    index = tabs.indexOf(container)
                    tabs.setTabIcon(index, icon)

    def iconSize(self):
        """ Get the icon size for the title bar.

        Returns
        -------
        result : QSize
            The size to use for the icons in the title bar.

        """
        return self.titleBarWidget().iconSize()

    def setIconSize(self, size):
        """ Set the icon size for the title bar.

        Parameters
        ----------
        icon : QSize
            The icon size to use for the title bar. Icons smaller than
            this size will not be scaled up.

        """
        self.titleBarWidget().setIconSize(size)

    def isLinked(self):
        """ Get whether or not this dock item is linked.

        Returns
        -------
        result : bool
            True if the item is linked, False otherwise.

        """
        return self.titleBarWidget().isLinked()

    def setLinked(self, linked):
        """ Set whether or not the dock item is linked.

        Parameters
        ----------
        linked : bool
            True if the dock item should be linked, False otherwise.

        """
        self.titleBarWidget().setLinked(linked)

    def isPinned(self):
        """ Get whether or not this dock item is pinned.

        Returns
        -------
        result : bool
            True if the item is pinned, False otherwise.

        """
        return self.titleBarWidget().isPinned()

    def setPinned(self, pinned, quiet=False):
        """ Set whether or not the dock item is pinned.

        Parameters
        ----------
        pinned : bool
            True if the dock item should be pinned, False otherwise.

        quiet : bool, optional
            True if the state should be set without emitted the toggled
            signal. The default is False.

        """
        self.titleBarWidget().setPinned(pinned, quiet)

    def isFloating(self):
        """ Get whether the dock item is free floating.

        """
        container = self.parent()
        if container is not None:
            return container.isWindow()
        return self.isWindow()

    def titleEditable(self):
        """ Get whether the title is user editable.

        Returns
        -------
        result : bool
            True if the title is user editable, False otherwise.

        """
        return self.titleBarWidget().isEditable()

    def setTitleEditable(self, editable):
        """ Set whether or not the title is user editable.

        Parameters
        ----------
        editable : bool
            True if the title is user editable, False otherwise.

        """
        self.titleBarWidget().setEditable(editable)

    def titleBarForceHidden(self):
        """ Get whether or not the title bar is force hidden.

        Returns
        -------
        result : bool
            Whether or not the title bar is force hidden.

        """
        return self.titleBarWidget().isForceHidden()

    def setTitleBarForceHidden(self, hidden):
        """ Set the force hidden state of the title bar.

        Parameters
        ----------
        hidden : bool
            True if the title bar should be hidden, False otherwise.

        """
        self.titleBarWidget().setForceHidden(hidden)

    def closable(self):
        """ Get whether or not the dock item is closable.

        Returns
        -------
        result : bool
            True if the dock item is closable, False otherwise.

        """
        return self._closable

    def setClosable(self, closable):
        """ Set whether or not the dock item is closable.

        Parameters
        ----------
        closable : bool
            True if the dock item is closable, False otherwise.

        """
        if closable != self._closable:
            self._closable = closable
            bar = self.titleBarWidget()
            buttons = bar.buttons()
            if closable:
                buttons |= bar.CloseButton
            else:
                buttons &= ~bar.CloseButton
            bar.setButtons(buttons)
            # A concession to practicality: walk the ancestry and update
            # the tab close button if this item lives in a dock tab.
            container = self.parent()
            if container is not None:
                stacked = container.parent()
                if stacked is not None:
                    tabs = stacked.parent()
                    if isinstance(tabs, QDockTabWidget):
                        index = tabs.indexOf(container)
                        tabs.setCloseButtonVisible(index, closable)

    def titleBarWidget(self):
        """ Get the title bar widget for the dock item.

        If a custom title bar has not been assigned, a default title
        bar will be returned. To prevent showing a title bar, set the
        visibility on the returned title bar to False.

        Returns
        -------
        result : IDockItemTitleBar
            An implementation of IDockItemTitleBar. This will never be
            None.

        """
        layout = self.layout()
        bar = layout.titleBarWidget()
        if bar is None:
            bar = QDockTitleBar()
            self.setTitleBarWidget(bar)
        return bar

    def setTitleBarWidget(self, title_bar):
        """ Set the title bar widget for the dock item.

        Parameters
        ----------
        title_bar : IDockItemTitleBar or None
            A custom implementation of IDockItemTitleBar, or None. If
            None, then the default title bar will be restored.

        """
        layout = self.layout()
        old = layout.titleBarWidget()
        if old is not None:
            old.maximizeButtonClicked.disconnect(self.maximizeButtonClicked)
            old.restoreButtonClicked.disconnect(self.restoreButtonClicked)
            old.closeButtonClicked.disconnect(self.closeButtonClicked)
            old.linkButtonToggled.disconnect(self.linkButtonToggled)
            old.pinButtonToggled.disconnect(self.pinButtonToggled)
            old.titleEdited.disconnect(self.titleEdited)
            old.leftDoubleClicked.disconnect(self.titleBarLeftDoubleClicked)
            old.rightClicked.disconnect(self.titleBarRightClicked)
        title_bar = title_bar or QDockTitleBar()
        title_bar.maximizeButtonClicked.connect(self.maximizeButtonClicked)
        title_bar.restoreButtonClicked.connect(self.restoreButtonClicked)
        title_bar.closeButtonClicked.connect(self.closeButtonClicked)
        title_bar.linkButtonToggled.connect(self.linkButtonToggled)
        title_bar.pinButtonToggled.connect(self.pinButtonToggled)
        title_bar.titleEdited.connect(self.titleEdited)
        title_bar.leftDoubleClicked.connect(self.titleBarLeftDoubleClicked)
        title_bar.rightClicked.connect(self.titleBarRightClicked)
        layout.setTitleBarWidget(title_bar)

    def dockWidget(self):
        """ Get the dock widget for this dock item.

        Returns
        -------
        result : QWidget or None
            The dock widget being managed by this item.

        """
        return self.layout().dockWidget()

    def setDockWidget(self, widget):
        """ Set the dock widget for this dock item.

        Parameters
        ----------
        widget : QWidget
            The QWidget to use as the dock widget in this item.

        """
        self.layout().setDockWidget(widget)

    def alert(self, level, on=250, off=250, repeat=4, persist=False):
        """ Set the alert level on the dock item.

        This will override any currently applied alert level.

        Parameters
        ----------
        level : unicode
            The alert level token to apply to the dock item.

        on : int
            The duration of the 'on' cycle, in ms. A value of -1 means
            always on.

        off : int
            The duration of the 'off' cycle, in ms. If 'on' is -1, this
            value is ignored.

        repeat : int
            The number of times to repeat the on-off cycle. If 'on' is
            -1, this value is ignored.

        persist : bool
            Whether to leave the alert in the 'on' state when the cycles
            finish. If 'on' is -1, this value is ignored.

        """
        if self._alert_data is not None:
            self.clearAlert()
        app = QApplication.instance()
        app.focusChanged.connect(self._onAppFocusChanged)
        timer = QTimer()
        timer.setSingleShot(True)
        timer.timeout.connect(self._onAlertTimer)
        on, off, repeat = max(-1, on), max(0, off), max(1, repeat)
        self._alert_data = _AlertData(timer, level, on, off, repeat, persist)
        if on < 0:
            self.alerted.emit(level)
        else:
            self._onAlertTimer()

    def clearAlert(self):
        """ Clear the current alert level, if any.

        """
        if self._alert_data is not None:
            self._alert_data.timer.stop()
            self._alert_data = None
            app = QApplication.instance()
            app.focusChanged.disconnect(self._onAppFocusChanged)
            self.alerted.emit(u'')

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _onAlertTimer(self):
        """ Handle the alert data timer timeout.

        This handler will refresh the alert level for the current tick,
        or clear|persist the alert level if the ticks have expired.

        """
        data = self._alert_data
        if data is not None:
            if not data.active:
                data.active = True
                data.timer.start(data.on)
                self.alerted.emit(data.level)
            else:
                data.active = False
                data.remaining -= 1
                if data.remaining > 0:
                    data.timer.start(data.off)
                    self.alerted.emit(u'')
                elif data.persist:
                    data.timer.stop()
                    self.alerted.emit(data.level)
                else:
                    self.clearAlert()

    def _onAlerted(self, level):
        """ A signal handler for the 'alerted' signal.

        This handler will set the 'alert' dynamic property on the
        dock item, the title bar, and the title bar label, and then
        repolish all three items.

        """
        level = level or None
        title_bar = self.titleBarWidget()
        label = title_bar.label()
        self.setProperty(u'alert', level)
        title_bar.setProperty(u'alert', level)
        label.setProperty(u'alert', level)
        repolish(label)
        repolish(title_bar)
        repolish(self)

    def _onAppFocusChanged(self, old, new):
        """ A signal handler for the 'focusChanged' app signal

        This handler will clear the alert if one of the descendant
        widgets or the item itself gains focus.

        """
        while new is not None:
            if new is self:
                self.clearAlert()
                break
            new = new.parent()

    def _onVisibilityTimer(self):
        """ Handle the visibility timer timeout.

        This handler will post the dock item visibility event to the
        root dock area.

        """
        area = self.rootDockArea()
        if area is not None and area.dockEventsEnabled():
            timer, visible = self._vis_changed
            evt_type = DockItemShown if visible else DockItemHidden
            event = QDockItemEvent(evt_type, self.objectName())
            QApplication.postEvent(area, event)
            self._vis_changed = None

    def _postVisibilityChange(self, visible):
        """ Post a visibility changed event for the dock item.

        This method collapses the post on a timer and will not emit
        the event when the visibility temporarily toggles bettwen
        states.

        Parameters
        ----------
        visible : bool
            True if the item was show, False if the item was hidden.

        """
        area = self.rootDockArea()
        if area is not None and area.dockEventsEnabled():
            vis_changed = self._vis_changed
            if vis_changed is None:
                timer = QTimer()
                timer.setSingleShot(True)
                timer.timeout.connect(self._onVisibilityTimer)
                self._vis_changed = (timer, visible)
                timer.start()
            else:
                timer, old_visible = vis_changed
                if old_visible != visible:
                    self._vis_changed = None
                    timer.stop()



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed

from enaml.widgets.mpl_canvas import ProxyMPLCanvas

from .QtCore import Qt, __version__ as QT_VERSION
from .QtWidgets import QFrame, QVBoxLayout

from .qt_control import QtControl

if QT_VERSION[0] == '4':
    from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg
    try:
        from matplotlib.backends.backend_qt4agg import (
            NavigationToolbar2QTAgg as NavigationToolbar2QT)
    except ImportError:
        from matplotlib.backends.backend_qt4agg import NavigationToolbar2QT
elif QT_VERSION[0] == '5':
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
    from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
else:

    raise RuntimeError("No known Matplotlib backend for qt version {} "
                       "(as reported by qtpy.QT_VERSION)".format(QT_VERSION))


class QtMPLCanvas(QtControl, ProxyMPLCanvas):
    """ A Qt implementation of an Enaml ProxyMPLCanvas.

    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QFrame)

    #--------------------------------------------------------------------------
    # Initialization API
    #--------------------------------------------------------------------------
    def create_widget(self):
        """ Create the underlying widget.

        """
        widget = QFrame(self.parent_widget())
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        widget.setLayout(layout)
        self.widget = widget

    def init_layout(self):
        """ Initialize the layout of the underlying widget.

        """
        super(QtMPLCanvas, self).init_layout()
        self._refresh_mpl_widget()

    #--------------------------------------------------------------------------
    # ProxyMPLCanvas API
    #--------------------------------------------------------------------------
    def set_figure(self, figure):
        """ Set the MPL figure for the widget.

        """
        with self.geometry_guard():
            self._refresh_mpl_widget()

    def set_toolbar_visible(self, visible):
        """ Set the toolbar visibility for the widget.

        """
        layout = self.widget.layout()
        if layout.count() == 2:
            with self.geometry_guard():
                toolbar = layout.itemAt(0).widget()
                toolbar.setVisible(visible)

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _refresh_mpl_widget(self):
        """ Create the mpl widget and update the underlying control.

        """
        # Delete the old widgets in the layout, it's just shenanigans
        # to try to reuse the old widgets when the figure changes.
        widget = self.widget
        layout = widget.layout()
        while layout.count():
            layout_item = layout.takeAt(0)
            layout_item.widget().deleteLater()

        # Create the new figure and toolbar widgets. It seems that key
        # events will not be processed without an mpl figure manager.
        # However, a figure manager will create a new toplevel window,
        # which is certainly not desired in this case. This appears to
        # be a limitation of matplotlib. The canvas is manually set to
        # visible, or QVBoxLayout will ignore it for size hinting.
        figure = self.declaration.figure
        if figure:
            canvas = FigureCanvasQTAgg(figure)
            canvas.setParent(widget)
            canvas.setFocusPolicy(Qt.ClickFocus)
            canvas.setVisible(True)
            toolbar = NavigationToolbar2QT(canvas, widget)
            toolbar.setVisible(self.declaration.toolbar_visible)
            layout.addWidget(toolbar)
            layout.addWidget(canvas)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from enaml.qt.QtCore import Qt
from enaml.qt.QtWidgets import QSplitter, QSplitterHandle

from .q_dock_area import QDockArea


class QDockSplitterHandle(QSplitterHandle):
    """ A subclass for dock splitter handles.

    This subclass sets the hover flag on the splitter handle so that
    the hover rules in a style sheet work correctly.

    """
    def __init__(self, orientation, parent):
        super(QDockSplitterHandle, self).__init__(orientation, parent)
        self.setAttribute(Qt.WA_Hover)


class QDockSplitter(QSplitter):
    """ A splitter subclass for distinguishing dock splitters.

    This subclass allows selecting the dock splitter in a stylesheet
    and otherwise distinguishing dock splitters from standard QSplitter
    instances used elsewhere in the application.

    """
    def createHandle(self):
        """ Create a sentinel dock splitter handle.

        """
        return QDockSplitterHandle(self.orientation(), self)

    def inheritOpaqueResize(self):
        """ Inherit the opaque resize state.

        This method traverses the ancestor hierarchy and updates its
        opaque resize state based on the first QDockArea it finds. If
        no such ancestor exists, the setting is unchanged. This method
        is called by the framework at the appropriate times.

        """
        p = self.parent()
        while p is not None:
            if isinstance(p, QDockArea):
                self.setOpaqueResize(p.opaqueItemResize())
                return
            p = p.parent()



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Typed

from enaml.widgets.datetime_selector import ProxyDatetimeSelector

from .QtWidgets import QDateTimeEdit

from .qt_bounded_datetime import QtBoundedDatetime, CHANGED_GUARD


class QtDatetimeSelector(QtBoundedDatetime, ProxyDatetimeSelector):
    """ A Qt implementation of an Enaml ProxyDatetimeSelector.

    """
    #: A reference to the widget created by the proxy.
    widget = Typed(QDateTimeEdit)

    #--------------------------------------------------------------------------
    # Initialization API
    #--------------------------------------------------------------------------
    def create_widget(self):
        """ Create the QDateTimeEdit widget.

        """
        self.widget = QDateTimeEdit(self.parent_widget())

    def init_widget(self):
        """ Initialize the widget.

        """
        super(QtDatetimeSelector, self).init_widget()
        d = self.declaration
        self.set_datetime_format(d.datetime_format)
        self.set_calendar_popup(d.calendar_popup)
        self.widget.dateTimeChanged.connect(self.on_datetime_changed)

    #--------------------------------------------------------------------------
    # Abstract API Implementation
    #--------------------------------------------------------------------------
    def get_datetime(self):
        """ Return the current datetime in the control.

        Returns
        -------
        result : datetime
            The current control datetime as a datetime object.

        """
        return self.widget.dateTime().toPython()

    def set_minimum(self, datetime):
        """ Set the widget's minimum datetime.

        Parameters
        ----------
        datetime : datetime
            The datetime object to use for setting the minimum datetime.

        """
        self.widget.setMinimumDateTime(datetime)

    def set_maximum(self, datetime):
        """ Set the widget's maximum datetime.

        Parameters
        ----------
        datetime : datetime
            The datetime object to use for setting the maximum datetime.

        """
        self.widget.setMaximumDateTime(datetime)

    def set_datetime(self, datetime):
        """ Set the widget's current datetime.

        Parameters
        ----------
        datetime : datetime
            The datetime object to use for setting the datetime.

        """
        self._guard |= CHANGED_GUARD
        try:
            self.widget.setDateTime(datetime)
        finally:
            self._guard &= ~CHANGED_GUARD

    def set_datetime_format(self, format):
        """ Set the widget's datetime format.

        Parameters
        ----------
        format : string
            A Python time formatting string.

        """
        # XXX make sure Python's and Qt's format strings are the
        # same, or convert between the two.
        self.widget.setDisplayFormat(format)

    def set_calendar_popup(self, popup):
        """ Set whether a calendar popup is available on the widget.

        Parameters
        ----------
        popup : bool
            Whether the calendar popup is enabled.

        """
        self.widget.setCalendarPopup(popup)



#------------------------------------------------------------------------------
# Copyright (c) 2013, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from atom.api import Bool, Typed, ForwardTyped, observe, set_default

from enaml.core.declarative import d_
from enaml.image import Image

from .control import Control, ProxyControl


class ProxyImageView(ProxyControl):
    """ The abstract definition of a proxy ImageView object.

    """
    #: A reference to the ImageView declaration.
    declaration = ForwardTyped(lambda: ImageView)

    def set_image(self, image):
        raise NotImplementedError

    def set_scale_to_fit(self, scale):
        raise NotImplementedError

    def set_allow_upscaling(self, allow):
        raise NotImplementedError

    def set_preserve_aspect_ratio(self, preserve):
        raise NotImplementedError

    def get_aspect_ratio(self):
        raise NotImplementedError


class ImageView(Control):
    """ A widget which can display an Image with optional scaling.

    """
    #: The image to display in the viewer.
    image = d_(Typed(Image))

    #: Whether or not to scale the image with the size of the component.
    scale_to_fit = d_(Bool(False))

    #: Whether to allow upscaling of an image if scale_to_fit is True.
    allow_upscaling = d_(Bool(True))

    #: Whether or not to preserve the aspect ratio if scaling the image.
    preserve_aspect_ratio = d_(Bool(True))

    #: An image view hugs its width weakly by default.
    hug_width = set_default('weak')

    #: An image view hugs its height weakly by default.
    hug_height = set_default('weak')

    #: A reference to the ProxyImageView object.
    proxy = Typed(ProxyImageView)

    def layout_constraints(self):
        """Add constraints to preserve the aspect ratio.

        """
        if self.proxy_is_active and self.preserve_aspect_ratio:
            ratio = self.proxy.get_aspect_ratio()
            return self.constraints + [self.width == ratio*self.height]
        return self.constraints

    #--------------------------------------------------------------------------
    # Observers
    #--------------------------------------------------------------------------
    @observe('image', 'scale_to_fit', 'allow_upscaling', 'preserve_aspect_ratio')
    def _update_proxy(self, change):
        """ An observer which sends state change to the proxy.

        """
        # The superclass handler implementation is sufficient.
        super(ImageView, self)._update_proxy(change)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from . import QT_API, PYSIDE_API, PYSIDE2_API, PYQT4_API, PYQT5_API
if QT_API in PYSIDE_API or QT_API in PYSIDE2_API:
    msg = 'the Qt Scintilla widget is only available when using PyQt'
    raise ImportError(msg)

import logging
import sys
import weakref

from atom.api import Typed

from enaml.colors import parse_color
from enaml.fonts import parse_font
from enaml.scintilla.scintilla import ProxyScintilla

if QT_API in PYQT4_API:
    from PyQt4 import Qsci
elif QT_API in PYQT5_API:
    from PyQt5 import Qsci
else:
    import QScintilla as Qsci

from .QtGui import QColor, QFont

from .q_resource_helpers import (QColor_from_Color, QFont_from_Font,
                                 get_cached_qimage)
from .qt_control import QtControl
from .scintilla_lexers import LEXERS, LEXERS_INV
from .scintilla_tokens import TOKENS


#: The module-level logger
logger = logging.getLogger(__name__)


Base = Qsci.QsciScintillaBase
QsciScintilla = Qsci.QsciScintilla


if sys.platform == 'win32':
    DEFAULT_EOL = 'crlf'
elif sys.platform == 'darwin':
    DEFAULT_EOL = 'cr'
else:
    DEFAULT_EOL = 'lf'


EOL_MODE = {
    'crlf': Base.SC_EOL_CRLF,
    'cr': Base.SC_EOL_CR,
    'lf': Base.SC_EOL_LF,
}


EDGE_MODE = {
    'none': Base.EDGE_NONE,
    'line': Base.EDGE_LINE,
    'background': Base.EDGE_BACKGROUND,
}


INDENTATION_GUIDES = {
    'none': Base.SC_IV_NONE,
    'real': Base.SC_IV_REAL,
    'look_forward': Base.SC_IV_LOOKFORWARD,
    'look_both': Base.SC_IV_LOOKBOTH,
}


WHITE_SPACE = {
    'visible_always': Base.SCWS_VISIBLEALWAYS,
    'visible_after_indent': Base.SCWS_VISIBLEAFTERINDENT,
    'invisible': Base.SCWS_INVISIBLE,
}

AUTOCOMPLETION_USE_SINGLE = {
    'never': QsciScintilla.AcusNever,
    'explicit': QsciScintilla.AcusExplicit,
    'always': QsciScintilla.AcusAlways,
}

AUTOCOMPLETION_SOURCE = {
    'none': QsciScintilla.AcsNone,
    'all': QsciScintilla.AcsAll,
    'document': QsciScintilla.AcsDocument,
    'apis': QsciScintilla.AcsAPIs,
}

INDICATOR_STYLE = {
    'plain': QsciScintilla.PlainIndicator,
    'squiggle': QsciScintilla.SquiggleIndicator,
    'tt': QsciScintilla.TTIndicator,
    'diagonal': QsciScintilla.DiagonalIndicator,
    'strike': QsciScintilla.StrikeIndicator,
    'hidden': QsciScintilla.HiddenIndicator,
    'box': QsciScintilla.BoxIndicator,
    'round_box': QsciScintilla.RoundBoxIndicator,
    'straight_box': QsciScintilla.StraightBoxIndicator,
    'full_box': QsciScintilla.FullBoxIndicator,
    'dashes': QsciScintilla.DashesIndicator,
    'dots': QsciScintilla.DotsIndicator,
    'squiggle_low': QsciScintilla.SquiggleLowIndicator,
    'dot_box': QsciScintilla.DotBoxIndicator,
    'thick_composition': QsciScintilla.ThickCompositionIndicator,
    'thin_composition': QsciScintilla.ThinCompositionIndicator,
    'text_color': QsciScintilla.TextColorIndicator,
    'triangle': QsciScintilla.TriangleIndicator,
    'triangle_character': QsciScintilla.TriangleCharacterIndicator,
}

NUMBER_MARGIN = 0


def _make_color(color_str):
    """ A function which converts a color string into a QColor.

    """
    color = parse_color(color_str)
    if color is not None:
        return QColor_from_Color(color)
    return QColor()


def _make_font(font_str):
    """ A function which converts a font string into a QColor.

    """
    font = parse_font(font_str)
    if font is not None:
        return QFont_from_Font(font)
    return QFont()


class QtScintilla(QtControl, ProxyScintilla):
    """ A Qt implementation of an Enaml ProxyScintilla.

    """
    #: A weak cache which maps uuid -> QsciDocument.
    qsci_doc_cache = weakref.WeakValueDictionary()

    #: A reference to the widget created by the proxy.
    widget = Typed(QsciScintilla)

    #: A reference to the autocomplete API
    qsci_api = Typed(Qsci.QsciAPIs)

    #: A strong reference to the QsciDocument handle.
    qsci_doc = Typed(Qsci.QsciDocument)

    #: Indicator style to style ID mapping
    _indicator_styles = Typed(dict, ())

    #: Marker image to marker ID mapping
    _marker_images = Typed(dict, ())

    #--------------------------------------------------------------------------
    # Initialization API
    #--------------------------------------------------------------------------
    def create_widget(self):
        """ Create the underlying label widget.

        """
        self.widget = QsciScintilla(self.parent_widget())

    def init_widget(self):
        """ Initialize the underlying widget.

        """
        super(QtScintilla, self).init_widget()
        d = self.declaration
        self.set_document(d.document)
        self.set_autocomplete(d.autocomplete)
        self.set_syntax(d.syntax, refresh_style=False)
        self.set_settings(d.settings)
        self.set_zoom(d.zoom)
        self.refresh_style()
        if d.indicators:
            self.set_indicators(d.indicators)
        if d.markers:
            self.set_markers(d.markers)
        self.widget.textChanged.connect(self.on_text_changed)
        self.widget.cursorPositionChanged.connect(
            self.on_cursor_position_changed)

    def destroy(self):
        """ A reimplemented destructor.

        This destructor decrefs its document handle before calling the
        superclass destructor. This prevents a segfault in PyQt.

        """
        # Clear the strong reference to the document. It must be freed
        # *before* the last widget using it is freed or PyQt segfaults.
        del self.qsci_doc
        if self.qsci_api:
            del self.qsci_api
        super(QtScintilla, self).destroy()

    #--------------------------------------------------------------------------
    # Signal Handlers
    #--------------------------------------------------------------------------
    def on_text_changed(self):
        """ Handle the 'textChanged' signal on the widget.

        """
        d = self.declaration
        if d is not None:
            d.text_changed()

            self.refresh_line_number_width()

    def on_cursor_position_changed(self):
        """ Handle the 'cursorPositionChanged' signal on the widget.

        """
        d = self.declaration
        if d is not None:
            d.cursor_position = self.widget.getCursorPosition()

    #--------------------------------------------------------------------------
    # Helper Methods
    #--------------------------------------------------------------------------
    def refresh_style(self):
        """ Refresh the theme styling for the widget.

        This method will style the widget and the lexer using the
        current theme that was provided by the declaration object.

        """
        colorcache = {}
        fontcache = {}

        def get_color(color_str):
            if color_str in colorcache:
                return colorcache[color_str]
            color = colorcache[color_str] = _make_color(color_str)
            return color

        def get_font(font_str):
            if font_str in fontcache:
                return fontcache[font_str]
            font = fontcache[font_str] = _make_font(font_str)
            return font

        def pull_color(dct, key, default):
            color = dct.get(key)
            if color is None:
                return default
            return get_color(color)

        def pull_font(dct, key, default):
            font = dct.get(key)
            if font is None:
                return default
            return get_font(font)

        # Setup the various defaults.
        caret_color = QColor(0, 0, 0)
        default_color = QColor(0, 0, 0)
        default_paper = QColor(255, 255, 255)
        default_font = QFont()

        # Update the defaults from the theme's root 'settings' object.
        theme = self.declaration.theme
        settings = theme.get('settings')
        if settings is not None:
            caret_color = pull_color(settings, 'caret', caret_color)
            default_color = pull_color(settings, 'color', default_color)
            default_paper = pull_color(settings, 'paper', default_paper)
            default_font = pull_font(settings, 'font', default_font)

        # Apply the default styling for the widget.
        widget = self.widget
        widget.setCaretForegroundColor(caret_color)
        widget.setColor(default_color)
        widget.setPaper(default_paper)
        widget.setFont(default_font)

        # Ensure the lexer and syntax tokens exist before continuing.
        lexer = widget.lexer()
        if lexer is None:
            return
        syntax = LEXERS_INV.get(type(lexer))
        if syntax is None:
            return

        # Resolve the default parameters and apply them for all lexer
        # styles. More specifc rules will override these defaults.
        syntax_rules = theme.get(syntax, {})
        syntax_tokens = TOKENS.get(syntax, {})
        syntax_default = syntax_rules.get('default', {})
        default_color = pull_color(syntax_default, 'color', default_color)
        default_paper = pull_color(syntax_default, 'paper', default_paper)
        default_font = pull_font(syntax_default, 'font', default_font)
        lexer.setColor(default_color)
        lexer.setPaper(default_paper)
        lexer.setFont(default_font)

        # Override the defaults with more specific syntax rules.
        for token, rule in syntax_rules.items():
            if token == 'default':
                continue
            qtoken = getattr(lexer, syntax_tokens.get(token, ''), None)
            if qtoken is not None:
                color = pull_color(rule, 'color', default_color)
                paper = pull_color(rule, 'paper', default_paper)
                font = pull_font(rule, 'font', default_font)
                lexer.setColor(color, qtoken)
                lexer.setPaper(paper, qtoken)
                lexer.setFont(font, qtoken)
            else:
                msg = "unknown token '%s' given for the '%s' syntax"
                logger.warn(msg % (token, syntax))

    def refresh_autocomplete(self):
        """ If the lexer changes, update the API options as these depend
        on the lexer.

        """
        d = self.declaration
        if d.autocomplete in ['api', 'all'] and d.autocompletions:
            self.set_autocompletions(d.autocompletions)

    def refresh_line_number_width(self):
        """ When the number of lines changes update the width accordingly.
        Always shows one more than the width of the number of lines.

        """
        w = self.widget
        # Only update it if show_line_numbers=True
        if w.marginWidth(NUMBER_MARGIN) > 0:
            w.setMarginWidth(NUMBER_MARGIN, "0"+str(max(10, w.lines())))

    def get_indicator_style_id(self, indicator):
        """ Get the indicator style id for this indicator. The key
        is simply the style and fg color.

        If the key does not exist, define a new style.

        """
        style = "{},{}".format(indicator.style, indicator.color)
        if style not in self._indicator_styles:
            w = self.widget
            style_id = w.indicatorDefine(INDICATOR_STYLE[indicator.style])
            w.setIndicatorForegroundColor(_make_color(indicator.color),
                                          style_id)
            self._indicator_styles[style] = style_id
        return self._indicator_styles[style]

    #--------------------------------------------------------------------------
    # ProxyScintilla API
    #--------------------------------------------------------------------------
    def set_document(self, document):
        """ Set the document on the underlying widget.

        """
        qdoc = self.qsci_doc_cache.get(document.uuid)
        if qdoc is None:
            qdoc = self.qsci_doc_cache[document.uuid] = Qsci.QsciDocument()
        self.qsci_doc = qdoc  # take a strong ref since PyQt doesn't
        self.widget.setDocument(qdoc)

    def set_syntax(self, syntax, refresh_style=True):
        """ Set the syntax on the underlying widget.

        """
        # The old lexer will remain as a child unless deleted.
        old = self.widget.lexer()
        if old is not None:
            old.deleteLater()
        lexer_cls = LEXERS.get(syntax) or (lambda w: None)
        self.widget.setLexer(lexer_cls(self.widget))
        self.refresh_autocomplete()
        if refresh_style:
            self.refresh_style()

    def set_theme(self, theme):
        """ Set the styling theme for the widget.

        """
        self.refresh_style()

    def set_settings(self, settings):
        """ Set the settings for the widget.

        """
        w = self.widget
        send = w.SendScintilla
        get = settings.get

        def pull(kind, name, default):
            value = get(name, default)
            if not isinstance(value, kind):
                msg = 'invalid value "%s" for "%s" setting'
                logger.warn(msg % (value, name))
                value = default
            return value

        pull_int = lambda name, default: pull(int, name, default)
        pull_bool = lambda name, default: pull(bool, name, default)
        pull_str = lambda name, default: pull(str, name, default)

        def pull_enum(options, name, default):
            value = pull_str(name, default)
            if value in options:
                return options[value]
            message = 'invalid value "%s" for "%s" setting'
            logger.warn(message % (value, name))
            return options[default]

        def pull_color(name, default):
            color = pull_str(name, default)
            return _make_color(color)

        # Line Endings
        send(w.SCI_SETEOLMODE, pull_enum(EOL_MODE, 'eol_mode', DEFAULT_EOL))
        send(w.SCI_SETVIEWEOL, pull_bool('view_eol', False))

        # Long Lines
        send(w.SCI_SETEDGEMODE, pull_enum(EDGE_MODE, 'edge_mode', 'none'))
        send(w.SCI_SETEDGECOLUMN, pull_int('edge_column', 79))
        send(w.SCI_SETEDGECOLOUR, pull_color('edge_color', ''))

        # Tabs and Indentation
        send(w.SCI_SETTABWIDTH, pull_int('tab_width', 8))
        send(w.SCI_SETUSETABS, pull_bool('use_tabs', True))
        send(w.SCI_SETINDENT, pull_int('indent', 0))
        send(w.SCI_SETTABINDENTS, pull_bool('tab_indents', False))
        send(w.SCI_SETBACKSPACEUNINDENTS,
            pull_bool('backspace_unindents', False))
        send(w.SCI_SETINDENTATIONGUIDES,
            pull_enum(INDENTATION_GUIDES, 'indentation_guides', 'none'))
        w.setAutoIndent(pull_bool('auto_indent', False))

        # White Space
        send(w.SCI_SETVIEWWS, pull_enum(WHITE_SPACE, 'view_ws', 'invisible'))
        send(w.SCI_SETWHITESPACESIZE, pull_int('white_space_size', 1))
        send(w.SCI_SETEXTRAASCENT, pull_int('extra_ascent', 0))
        send(w.SCI_SETEXTRADESCENT, pull_int('extra_descent', 0))

        # Autocompletion
        # See https://qscintilla.com/general-autocompletion/
        w.setAutoCompletionThreshold(pull_int('autocompletion_threshold', 3))
        w.setAutoCompletionCaseSensitivity(
            pull_bool('autocompletion_case_sensitive', False))
        w.setAutoCompletionReplaceWord(
            pull_bool('autocompletion_replace_word', False))
        w.setAutoCompletionUseSingle(
            pull_enum(AUTOCOMPLETION_USE_SINGLE,
                      'autocompletion_use_single', 'never'))
        self.set_autocompletion_images(
            settings.get('autocompletion_images', []))

        # Line numbers
        self.set_show_line_numbers(pull_bool('show_line_numbers', False))

    def set_zoom(self, zoom):
        """ Set the zoom factor on the widget.

        """
        self.widget.zoomTo(zoom)

    def get_text(self):
        """ Get the text in the document.

        """
        return self.widget.text()

    def set_text(self, text):
        """ Set the text in the document.

        """
        self.widget.setText(text)

    def set_autocomplete(self, mode):
        """ Set the autocompletion mode

        """
        self.widget.setAutoCompletionSource(AUTOCOMPLETION_SOURCE[mode])

    def set_autocompletions(self, options):
        """ Set the autocompletion options for when the autocompletion mode
        is in 'all' or 'apis'.

        """
        # Delete the old if one exists
        if self.qsci_api:
            # Please note that it is not possible to add or remove entries
            # once you’ve “prepared” so we have to destroy and create
            # a new provider every time.
            self.qsci_api.deleteLater()
            self.qsci_api = None

        # Add the new options
        api = self.qsci_api = Qsci.QsciAPIs(self.widget.lexer())
        for option in options:
            api.add(option)
        api.prepare()

    def set_autocompletion_images(self, images):
        """ Set the images that can be used in autocompletion results.

        """
        w = self.widget
        w.clearRegisteredImages()
        for i, image in enumerate(images):
            w.registerImage(i, get_cached_qimage(image))

    def set_show_line_numbers(self, show):
        """ Set whether line numbers are shown or not by setting
        the margin width of the LineNumber margin.

        """
        w = self.widget
        w.setMarginType(NUMBER_MARGIN, QsciScintilla.NumberMargin)
        self.widget.setMarginWidth(
            NUMBER_MARGIN, "0"+str(max(10, w.lines())) if show else "")

    def set_markers(self, markers):
        """ Set the markers on the left margin of the widget.

        If the image is not a defined marker, one will be created.

        """
        w = self.widget

        # Clear markers
        w.markerDeleteAll()

        # Add the new markers
        for m in markers:
            # Define a new marker with the given image if one has not already
            # been created.
            if m.image not in self._marker_images:
                self._marker_images[m.image] = w.markerDefine(
                    get_cached_qimage(m.image))

            # Add the marker
            w.markerAdd(m.line, self._marker_images[m.image])

    def set_indicators(self, indicators):
        """ Set the indicators of the widget.

        This lets certain text be highlighted or underlined with a given
        style to indicate something (errors) within the editor.

        """
        w = self.widget

        # Cleanup old indicators by clearing all indicators in the document
        # There's no api to do this so clear the entire document range
        # for each style to ensure a clean state.
        lines = w.lines()
        column = w.lineLength(lines)
        for style_id in self._indicator_styles.values():
            w.clearIndicatorRange(0, 0, lines, column, style_id)

        # Add new indicators
        for ind in indicators:
            l0, c0 = ind.start
            l1, c1 = ind.stop
            w.fillIndicatorRange(l0, c0, l1, c1,
                                 self.get_indicator_style_id(ind))

    #--------------------------------------------------------------------------
    # Reimplementations
    #--------------------------------------------------------------------------
    def set_foreground(self, foreground):
        """ Set the foreground color of the widget.

        This reimplementation ignores the foreground setting. The
        foreground color is set by the theme.

        """
        pass

    def set_background(self, background):
        """ Set the background color of the widget.

        This reimplementation ignores the background setting. The
        background color is set by the theme.

        """
        pass

    def set_font(self, font):
        """ Set the font of the widget.

        This reimplementation ignores the font setting. The font is
        set by the theme.

        """
        pass
